(* Forward tcp connections over a connection with defined lag.
   Could be extended to support udp.. *)
(* (C) Erkki Seppala 2007-2008 *)

let debug = ref false

let bw_sample_history = 5.0		(* keep samples at most n seconds old *)

let create_server_socket port =
  let s = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
    try
      Unix.setsockopt s Unix.SO_REUSEADDR true;
      Unix.bind s (Unix.ADDR_INET (Unix.inet_addr_any, port));
      (try Unix.bind s (Unix.ADDR_INET (Unix.inet6_addr_any, port)) with exn -> ());
      Unix.listen s 0;
      s
    with exn ->
      Unix.close s;
      raise exn

let create_client_socket address =
  let s = 
    Unix.socket 
      (match address with 
	 | Unix.ADDR_UNIX _ -> Unix.PF_UNIX 
	 | Unix.ADDR_INET (addr, _) when try ignore (String.index (Unix.string_of_inet_addr addr) ':'); true with _ -> false -> Unix.PF_INET6
	 | Unix.ADDR_INET (addr, _) -> Unix.PF_INET)
	 Unix.SOCK_STREAM 0 in
    try
      Unix.connect s address;
      s
    with exn ->
      Unix.close s;
      raise exn 

let select rd wr ex sleep =
  let sleep =
    match sleep with 
      | Some x when x <= 0.0 -> 0.0
      | Some x -> x
      | None -> -1.0
  in
    Unix.select rd wr ex sleep

exception Eof

let read sock str ofs len =
  try
    let n = Unix.read sock str ofs len in
      if n = 0 then raise Eof
      else Some n
  with
    | Unix.Unix_error ((Unix.EAGAIN | Unix.EINTR), _, _) ->
	None

let write sock str ofs len =
  try
    let n = Unix.write sock str ofs len in
      if n = 0 then raise Eof
      else Some n
  with
    | Unix.Unix_error ((Unix.EAGAIN | Unix.EINTR), _, _) ->
	None
    | Unix.Unix_error ((Unix.EPIPE), _, _) ->
        raise Eof

let or_set (rd1, wr1, ex1) (rd2, wr2, ex2) =
  (rd1 @ rd2, wr1 @ wr2, ex1 @ ex2)

module TimeAmountSet = Set.Make (struct type t = (float * int) let compare = compare end)
module TAS = TimeAmountSet

type delay = { src	: Unix.file_descr;
	       dst	: Unix.file_descr;
	       rd_queue	: (float * string) Queue.t;
	       wr_queue : string Queue.t;
	       mutable partial  : (string * int) option;
	       delay	: float;
	       mutable wr_samples : TAS.t;
	       mutable wr_sum : int;
	       bw_limit : float option;
	       buf	: string;
	       mutable eof : bool;
	       finished : bool ref;
	       peer_finished : bool ref;
	     }

let new_delay src dst delay bw_limit finished peer_finished =
  { src = src; 
    dst = dst; 
    rd_queue = Queue.create (); 
    wr_queue = Queue.create (); 
    delay = delay;
    bw_limit = bw_limit;
    buf = String.make 1024 ' ';
    wr_samples = TAS.empty;
    wr_sum = 0;
    partial = None; 
    eof = false;
    finished = finished;
    peer_finished = peer_finished;
  }

let bw_delta d =
  if not (TAS.is_empty d.wr_samples) then
    fst (TAS.max_elt d.wr_samples) -. fst (TAS.min_elt d.wr_samples)
  else
    0.0

let calc_bw d =
  if not (TAS.is_empty d.wr_samples) then
    let delta = bw_delta d in
      if delta >= 0.01 then
	float d.wr_sum /. delta
      else
	0.0
  else
    0.0

let sample d now bytes =
  match d.bw_limit with
    | None -> () (* don't bother sampling *)
    | Some _ ->
	d.wr_samples <- TAS.add (now, bytes) d.wr_samples;
	d.wr_sum <- d.wr_sum + bytes;
	if !debug then Printf.printf "Adding sample of %d bytes at %f\n%!" bytes now;
	while not (TAS.is_empty d.wr_samples) && fst (TAS.min_elt d.wr_samples) < now -. bw_sample_history do
	  let (time, bytes) as elt = TAS.min_elt d.wr_samples in
	    if !debug then Printf.printf "Removing sample of %d bytes at %f\n%!" bytes time;
	    d.wr_samples <- TAS.remove elt d.wr_samples;
	    d.wr_sum <- d.wr_sum - bytes;
	done

let send_delay d =
  match d.bw_limit with
    | None -> d.delay
    | Some limit ->
	let bw = calc_bw d in
	  if !debug then Printf.printf "bw=%f delta=%f\n%!" bw (bw_delta d);
	  if bw > limit then
	    let next_slot = float d.wr_sum /. limit -. bw_delta d in
	      if !debug then Printf.printf "Packet delayed by %f\n%!" next_slot;
	      d.delay +. next_slot
	  else
	    d.delay

let finish d =
  d.finished := true;
  if !(d.peer_finished) then
    begin
      Unix.close d.src;
      Unix.close d.dst;
    end

let select_set d now =
  if !(d.peer_finished) && Queue.is_empty d.wr_queue && d.partial == None then
    begin
      finish d;
      (([], [], []), None)
    end
  else
    let rd = if d.eof then [] else [d.src] in
    let wr = if Queue.is_empty d.wr_queue && d.partial = None then [] else [d.dst] in
      if !debug then Printf.printf "%d->%d now=%f\n%!" (Obj.magic d.src : int) (Obj.magic d.dst : int) now;
      match try Some (Queue.peek d.rd_queue) with Queue.Empty -> None with
	| Some (x, _) when x <= now -> 
            if !debug then Printf.printf "1: %f\n%!" x;
	    ((rd, [d.dst], []), Some (0.0))
	| Some (x, _) ->
            if !debug then Printf.printf "2: %f\n%!" x;
   	    ((rd, wr, []), Some (x -. now))
	| None -> 
            if !debug then Printf.printf "3\n%!";
            ((rd, wr, []), None)

let check d now (rd, wr, _) =
  let rec check_times_loop () =
    match try Some (Queue.peek d.rd_queue) with Queue.Empty -> None with
      | Some (x, _) when x <= now -> 
          if !debug then Printf.printf "Moving %d->%d\n%!" (Obj.magic d.src : int) (Obj.magic d.dst : int);
	  Queue.add (snd (Queue.take d.rd_queue)) d.wr_queue; 
	  check_times_loop ()
      | _ -> ()
  in
    check_times_loop ();
    if !debug then Printf.printf "%d->%d wr_queue data? %b partial? %b\n%!" (Obj.magic d.src) (Obj.magic d.dst) (not (Queue.is_empty d.wr_queue)) (d.partial <> None);
    if not (Queue.is_empty d.wr_queue) && d.partial = None then
      d.partial <- Some (Queue.take d.wr_queue, 0);
    begin
      match d.partial with
	| Some (msg, ofs) when List.mem d.dst wr ->
	    begin
	      (* handle write errors.. *)
	      try
	        match write d.dst msg ofs (String.length msg - ofs) with
		  | Some n -> 
		      if ofs + n = String.length msg then
		        begin
		          d.partial <- None;
		          if d.eof then
   	                    finish d;
                        end
		      else
		        d.partial <- Some (msg, ofs + n)
		  | None -> ()
              with Eof -> finish d
	    end
	| _ -> ()
    end;
    if List.mem d.src rd && not d.eof then
      begin
        try
  	  match read d.src d.buf 0 (String.length d.buf) with
	    | Some n -> 
		if !debug then Printf.printf "Read %d->%d\n%!" (Obj.magic d.src : int) (Obj.magic d.dst : int);
		let time = now +. send_delay d in
		  Queue.add (time, String.sub d.buf 0 n) d.rd_queue;
		  sample d time (String.length d.buf);
	    | None -> ()
        with Eof -> 
          if !debug then Printf.printf "%d->%d Setting eof\n%!" (Obj.magic d.src) (Obj.magic d.dst);
          d.eof <- true
      end


let main src (dsthost, dst) delay bw_limit =
  let serv_fd = create_server_socket src in

  let rec loop delays =
    let now = Unix.gettimeofday () in
    if !debug then Printf.printf "Select sets\n%!";
    let ((rd, wr, ex), sleep) =
      List.fold_left 
	(fun (set, sleep) delay ->
	   let set', sleep' = select_set delay now in
	     (or_set set set',
	      match sleep, sleep' with
		| None, _ -> sleep'
		| Some old, Some new' when new' < old -> sleep'
		| _, _ -> sleep)
	)
	(([], [], []), None)
	delays
    in
    if !debug then Printf.printf "End select sets\n%!";
    let ((rd, _, _) as set) = select (serv_fd::rd) wr ex sleep in
    let now = Unix.gettimeofday () in
    let delays =
      if List.mem serv_fd rd then
        let client_fd, _ = Unix.accept serv_fd in
        let conn_fd = create_client_socket (Unix.ADDR_INET ((Unix.gethostbyname dsthost).Unix.h_addr_list.(0), dst)) in
        let () = Unix.set_nonblock client_fd in
        let () = Unix.set_nonblock conn_fd in
	let finish1 = ref false in
	let finish2 = ref false in
        let d1 = new_delay client_fd conn_fd delay bw_limit finish1 finish2 in
        let d2 = new_delay conn_fd client_fd delay bw_limit finish2 finish1 in
          if !debug then Printf.printf "client delay: %d->%d\n" (Obj.magic client_fd : int) (Obj.magic conn_fd : int);
          if !debug then Printf.printf "server delay: %d->%d\n" (Obj.magic conn_fd) (Obj.magic client_fd);
          d1::d2::delays
      else 
        delays
    in
      List.iter (fun d -> check d now set) delays;
      loop (List.filter (fun d -> not !(d.finished)) delays)
  in
    loop []

let usage =
  "lfwd - forward tcp traffic with a certain latency"

let _ = 
  let _ = Sys.signal Sys.sigpipe Sys.Signal_ignore in
  let src_port = ref None in
  let dst_port = ref None in
  let dst_host = ref "localhost" in
  let delay = ref 0.0 in
  let bw_limit = ref None in
    Arg.parse ["--sport", (Arg.Int (fun i -> src_port := Some i)), "Source port";
               "--dhost", (Arg.String (fun s -> dst_host := s)), "Destination host (localhost by default)";
	       "--dport", (Arg.Int (fun i -> dst_port := Some i)), "Destination port";
 	       "--delay", (Arg.Float (fun f -> delay := f)), "Traffic delay in seconds";
 	       "--bwlimit", (Arg.Float (fun f -> bw_limit := Some f)), "Bandwidth limit in bytes per second";
 	       "--debug", (Arg.Set debug), "Enable debug output"
	      ] 
      (fun _ -> failwith "Invalid parameter")
      usage;
    match !src_port, !dst_port with
      | Some src_port, Some dst_port ->
	  main src_port (!dst_host, dst_port) !delay !bw_limit
      | _ -> Printf.printf "Source port (--sport port) or destination port (--dport port) missing - try --help\n"

